const fs = require ("fs"); 
const utils = require ("daveutils"); 
const feedtools = require ("davereader"); 
const request = require ("request");
const electron = require ("electron");

var config = {
	productname: "electricRiver",
	productnameForDisplay: "Electric River",
	description: "River5 in an Electron shell.",
	userDataFolder: undefined,
	flOpenDevToolsAtStart: false,
	iconPath: "river5.png",
	urlInitialSubscriptionList: "http://rss2.io/code/electron/lists/initialsubscriptionlist.opml",
	fnameInitialSubscriptionList: "mySubs.opml",
	maxRiverItems: 300,
	homePage: {
		"panels": [
			{
				"title": "My River",
				"river": "mySubs.js"
				}
			]
		},
	flBuildEveryFiveSeconds: true, //3/29/17 by DW
	flDownloadPodcasts: true, //4/20/17 by DW
	version: "0.42a"
	};
var fnameConfig = "config.json";

var electronStats = {
	ctStarts: 0,
	ctSeconds: 0,
	ctStatsWrites: 0
	};
var fnameStats = "data/electronStats.json", flStatsChanged = false;

var mainWindow = null; 

var bullMancuso = undefined;

function debugMessage (s) {
	console.log (s);
	try {
		if (bullMancuso !== undefined) {
			bullMancuso.send ("debugMessage", s);
			}
		}
	catch (err) {
		}
	}
function statsChanged () {
	flStatsChanged = true;
	}
function fsSureFilePathSync (path) { 
	var splits = path.split ("/");
	path = ""; 
	if (splits.length > 0) {
		function doLevel (levelnum) {
			if (levelnum < (splits.length - 1)) {
				path += splits [levelnum] + "/";
				if (fs.existsSync (path)) {
					doLevel (levelnum + 1);
					}
				else {
					fs.mkdirSync (path, undefined);
					doLevel (levelnum + 1);
					}
				}
			}
		doLevel (0);
		}
	return (true);
	}
function fsSureFilePath (path, callback) { 
	var splits = path.split ("/");
	path = ""; //1/8/15 by DW
	if (splits.length > 0) {
		function doLevel (levelnum) {
			if (levelnum < (splits.length - 1)) {
				path += splits [levelnum] + "/";
				fs.exists (path, function (flExists) {
					if (flExists) {
						doLevel (levelnum + 1);
						}
					else {
						fs.mkdir (path, undefined, function () {
							doLevel (levelnum + 1);
							});
						}
					});
				}
			else {
				if (callback != undefined) {
					callback ();
					}
				}
			}
		doLevel (0);
		}
	else {
		if (callback != undefined) {
			callback ();
			}
		}
	}
function fsSureFolder (folder, callback) {
	fsSureFilePath (folder + "x", callback);
	return (folder);
	}
function fsIsFolder (path) {
	return (fs.statSync (path).isDirectory ());
	}
function feedToolsRequest (method, path, callback) { //send an HTTP request directly to River5, without going through network
	console.log ("feedToolsRequest: method == " + method + ", path == " + path);
	var request = {
		headers: {
			host: "localhost:1337"
			},
		method: method,
		url: path
		};
	var response = {
		code: undefined,
		headers: {
			},
		writeHead: function (code, headers) {
			this.code = code;
			this.headers = headers;
			},
		end: function (val) {
			if (callback !== undefined) {
				callback (this.headers, val);
				}
			}
		};
	feedtools.httpRequest (request, response);
	}
function electronStartup () {
	electron.powerSaveBlocker.start ("prevent-app-suspension");
	electron.app.on ("ready", function () { 
		mainWindow = new electron.BrowserWindow ({
			width: 1100, 
			height: 1000,
			icon: __dirname + "/" + config.iconPath //3/17/17 by DW
			});
		mainWindow.on ("closed", function () {
			mainWindow = null;
			});
		mainWindow.webContents.on ("will-navigate", function (event, url) { 
			electron.shell.openExternal (url);
			event.preventDefault ();
			});
		});
	electron.app.on ("window-all-closed", function () { 
		electron.app.quit ();
		});
	electron.ipcMain.on ("asynch-message", function (event, arg1, arg2, arg3) {
		debugMessage ("IPC message received == " + arg1 + ", arg2 == " + arg2 + ", arg3 == " + arg3);
		switch (arg1) {
			case "hello":
				bullMancuso = event.sender;
				debugMessage ("hello message received");
				bullMancuso.send ("config", utils.jsonStringify (config));
				break;
			case "runScript":
				var val;
				debugMessage ("runScript: " + arg2);
				try {
					val = runUserScript (arg2, "quickscript");
					debugMessage ("runScript: val == " + val);
					event.sender.send ("scriptReturnVal", val);
					}
				catch (err) {
					debugMessage ("runScript: err.message == " + err.message);
					event.sender.send ("scriptReturnVal", undefined, err.message);
					}
				break;
			case "openUrl": 
				electron.shell.openExternal (arg2);
				break;
			case "openFolder": 
				electron.shell.showItemInFolder (arg2);
				break;
			case "httpRequest":
				feedToolsRequest (arg2, arg3, function (headers, val) {
					event.sender.send ("httpReturnVal", headers, val);
					});
				break;
			case "readAllFeedsNow": //4/18/17 by DW
				feedtools.readAllFeedsNow ();
				break;
			case "setConfig": //4/21/17 by DW
				var theConfigSubset = JSON.parse (arg2);
				for (x in theConfigSubset) {
					config [x] = theConfigSubset [x];
					}
				updateConfigJson (theConfigSubset);
				break;
			}
		});
	electron.ipcMain.on ("synchronous-message", function (event, arg) {
		debugMessage ("IPC message received == " + arg);
		switch (arg) {
			case "waitForOAuth":
				flOpenUrlsInExternalBrowser = false; //4/17/16 by DW
				if (myHttpServer === undefined) { //4/15/16 by DW -- can't do this twice
					myHttpServer = require ("http").createServer (tinyHttpServer);
					myHttpServer.listen (1401);
					}
				event.returnValue = 1401;
				break;
			case "getConfig": //4/21/17 by DW
				event.returnValue = utils.jsonStringify (config);
				break;
			}
		});
	}
function electronOpenHomePage () {
	feedToolsRequest ("GET", "/", function (headers, htmltext) {
		var findex = config.userDataFolder + "index.html";
		fs.writeFile (findex, htmltext, function (err) {
			if (err) {
				debugMessage ("electronStart: error writing index file " + findex + " == " + err.message);
				}
			else {
				if (config.flOpenDevToolsAtStart) {
					mainWindow.webContents.openDevTools ();
					}
				mainWindow.loadURL ("file://" + findex);
				}
			});
		});
	}
function getLocalFilePath (relpath) {
	return (config.userDataFolder + "/" + relpath);
	}
function writeConfig (f, config) {
	fsSureFilePath (f, function () {
		fs.writeFile (f, utils.jsonStringify (config), function (err) {
			if (err) {
				debugMessage ("writeConfig: error writing file " + f + " == " + err.message);
				}
			});
		});
	}
function readConfig (f, config, callback) {
	console.log ("readConfig: f == " + f); 
	fsSureFilePath (f, function () {
		fs.readFile (f, function (err, data) {
			if (!err) {
				try {
					var jstruct = JSON.parse (data.toString ());
					console.log (utils.jsonStringify (jstruct)); 
					for (var x in jstruct) {
						config [x] = jstruct [x];
						}
					}
				catch (err) {
					debugMessage ("readConfig: err == " + err.message);
					}
				}
			if (callback !== undefined) {
				callback ();
				}
			});
		});
	}
function writeUserDataFile (path, config) {
	var f = getLocalFilePath (path);
	fsSureFilePath (f, function () {
		fs.writeFile (f, utils.jsonStringify (config), function (err) {
			if (err) {
				debugMessage ("writeUserDataFile: error writing file " + f + " == " + err.message);
				}
			});
		});
	}
function readUserDataFile (path, config, callback) {
	var f = getLocalFilePath (path);
	debugMessage ("readUserDataFile: f == " + f);
	fsSureFilePath (f, function () {
		fs.readFile (f, function (err, data) {
			if (!err) {
				try {
					var jstruct = JSON.parse (data.toString ());
					for (var x in jstruct) {
						config [x] = jstruct [x];
						}
					}
				catch (err) {
					debugMessage ("readUserDataFile: err == " + err.message);
					}
				}
			if (callback !== undefined) {
				callback ();
				}
			});
		});
	}
function sureConfigJson (callback) { //4/18/17 by DW
	var f = getLocalFilePath (fnameConfig);
	var initConfig = {
		listsFolder: config.listsFolder,
		dataFolder: config.dataFolder,
		riversFolder: config.riversFolder,
		podcastsFolder: config.podcastsFolder,
		flDownloadPodcasts: config.flDownloadPodcasts, //4/20/17 by DW
		flOpenDevToolsAtStart: config.flOpenDevToolsAtStart,
		maxRiverItems: config.maxRiverItems,
		homePage: config.homePage
		};
	fsSureFilePath (f, function () {
		fs.readFile (f, function (err, data) {
			if (err) {
				try {
					fs.writeFile (f, utils.jsonStringify (initConfig), function (err) {
						});
					}
				catch (err) {
					debugMessage ("sureConfigJson: err == " + err.message);
					}
				}
			if (callback !== undefined) {
				callback ();
				}
			});
		});
	}
function updateConfigJson (theConfigSubset, callback) { //4/21/17 by DW
	var f = getLocalFilePath (fnameConfig);
	fsSureFilePath (f, function () {
		fs.readFile (f, function (err, data) {
			var theFullConfig = new Object ();
			if (!err) {
				try {
					theFullConfig = JSON.parse (data.toString ());
					}
				catch (err) {
					}
				}
			for (var x in theConfigSubset) {
				theFullConfig [x] = theConfigSubset [x]; 
				}
			fs.writeFile (f, utils.jsonStringify (theFullConfig), function (err) {
				});
			if (callback !== undefined) {
				callback ();
				}
			});
		});
	}
function electronEverySecond () {
	var now = new Date ();
	electronStats.ctSeconds++; 
	electronStats.whenLastEverySecond = now;
	if (flStatsChanged) {
		flStatsChanged = false;
		electronStats.ctStatsWrites++;
		electronStats.whenLastStatsWrite = now;
		writeUserDataFile (fnameStats, electronStats);
		}
	}
function copyInitialLists () {
	var sourcefolder = "lists/";
	var destfolder = config.listsFolder;
	fsSureFolder (destfolder, function () {
		fs.readdir (sourcefolder, function (err, list) {
			if (list !== undefined) { 
				for (var i = 0; i < list.length; i++) {
					var fname = list [i], fsource = sourcefolder + fname, fdest = destfolder + fname;
					try {
						fs.writeFileSync (fdest, fs.readFileSync (fsource))
						}
					catch (err) {
						debugMessage ("copyInitialLists: error copying \"" + fdest + "\" == " + err.message);
						}
					}
				}
			});
		});
	}
function downloadInitialList (callback) {
	request (config.urlInitialSubscriptionList, function (error, response, body) {
		if (!error && (response.statusCode == 200)) {
			var fdest = config.listsFolder + config.fnameInitialSubscriptionList;
			fsSureFilePath (fdest, function () {
				fs.writeFile (fdest, body, function (error, response, body) {
					if (callback !== undefined) {
						callback ();
						}
					});
				});
			}
		});
	}
function startup () {
	var now = new Date ();
	console.log ("startup");
	electronStartup ();
	config.userDataFolder = electron.app.getPath ("userData") + "/river5/"; 
	readConfig (fnameConfig, config, function () {
		var basefolder = config.userDataFolder;
		fsSureFolder (basefolder, function () {
			config.listsFolder = fsSureFolder (basefolder + "lists/");
			config.dataFolder = fsSureFolder (basefolder + "data/");
			config.riversFolder = fsSureFolder (basefolder + "rivers/");
			config.podcastsFolder = fsSureFolder (basefolder + "podcasts/"); //4/17/17 by DW
			config.addToRiverCallbacksFolder = fsSureFolder (basefolder + "callbacks/addToRiver/"); //4/16/17 by DW
			console.log ("startup: config.addToRiverCallbacksFolder == " + config.addToRiverCallbacksFolder);
			config.flRequestCloudNotify = false;
			config.urlServerHomePageSource = "http://rss2.io/code/electron/index.html";
			config.flHttpEnabled = false;
			config.flWebSocketEnabled = false;
			config.notifyListenersCallback = function (s) {
				if (bullMancuso !== undefined) {
					bullMancuso.send ("webSocketCallback", s);
					}
				};
			config.statsChangedCallback = function (jsontext) {
				if (bullMancuso !== undefined) {
					bullMancuso.send ("statsChangedCallback", jsontext);
					}
				};
			config.consoleLogCallback = function (s) {
				if (bullMancuso !== undefined) {
					bullMancuso.send ("debugMessage", s);
					}
				};
			readUserDataFile (fnameConfig, config, function () { //3/29/17 by DW
				console.log ("startup config == " + utils.jsonStringify (config));
				readUserDataFile (fnameStats, electronStats, function () {
					function finishStartup () {
						electronStats.ctStarts++;
						electronStats.whenLastStart = now;
						statsChanged ();
						sureConfigJson (); //4/18/17 by DW -- make sure there's a config.json in the user's data directory
						feedtools.init (config, function () {
							electronOpenHomePage ();
							setInterval (electronEverySecond, 1000); 
							});
						}
					if (electronStats.ctStarts == 0) {
						downloadInitialList (finishStartup);
						}
					else {
						finishStartup ();
						}
					});
				});
			});
		});
	}
startup ();
